; Copyright (C) 1998  Transarc Corporation.
; All rights reserved.

; AfsShlExt.def : Declares the module parameters for the DLL.

LIBRARY      "afs_shl_ext"
DESCRIPTION  'AfsShlExt Windows Dynamic Link Library'

EXPORTS
    ; Explicit exports can go here
	DllCanUnloadNow PRIVATE
	DllGetClassObject PRIVATE
	DllRegisterServer PRIVATE
